@@ -1,5 +1,3 @@
-from __future__ import annotations
-
 from copy import copy
 from importlib import resources
 from io import BytesIO, TextIOWrapper
@@ -11,7 +9,7 @@ import sys
 from typing import IO, BinaryIO, Literal
 
 from configobj import ConfigObj, ConfigObjError
-import pyaes
+from Cryptodome.Cipher import AES
 
 logger = logging.getLogger(__name__)
 
@@ -25,7 +23,7 @@ def log(logger: logging.Logger, level: int, message: str) -> None:
     logger.log(level, message)
 
 
-def read_config_file(f: str | TextIOWrapper, list_values: bool = True) -> ConfigObj | None:
+def read_config_file(f: str | IO[str], list_values: bool = True) -> ConfigObj | None:
     """Read a config file.
 
     *list_values* set to `True` is the default behavior of ConfigObj.
@@ -52,7 +50,7 @@ def read_config_file(f: str | TextIOWrapper, list_values: bool = True) -> Config
     return config
 
 
-def get_included_configs(config_file: str | TextIOWrapper) -> list[str]:
+def get_included_configs(config_file: str | IO[str]) -> list[str | IO[str]]:
     """Get a list of configuration files that are included into config_path
     with !includedir directive.
 
@@ -64,7 +62,7 @@ def get_included_configs(config_file: str | TextIOWrapper) -> list[str]:
     """
     if not isinstance(config_file, str) or not os.path.isfile(config_file):
         return []
-    included_configs = []
+    included_configs: list[str | IO[str]] = []
 
     try:
         with open(config_file) as f:
@@ -80,10 +78,22 @@ def get_included_configs(config_file: str | TextIOWrapper) -> list[str]:
     return included_configs
 
 
-def read_config_files(files: list[str], list_values: bool = True) -> ConfigObj:
+def read_config_files(
+    files: list[str | IO[str]],
+    list_values: bool = True,
+    ignore_package_defaults: bool = False,
+    ignore_user_options: bool = False,
+) -> ConfigObj:
     """Read and merge a list of config files."""
 
-    config = create_default_config(list_values=list_values)
+    if ignore_package_defaults:
+        config = ConfigObj()
+    else:
+        config = create_default_config(list_values=list_values)
+
+    if ignore_user_options:
+        return config
+
     _files = copy(files)
     while _files:
         _file = _files.pop(0)
@@ -103,14 +113,15 @@ def read_config_files(files: list[str], list_values: bool = True) -> ConfigObj:
 def create_default_config(list_values: bool = True) -> ConfigObj:
     import mycli
 
-    default_config_file = resources.open_text(mycli, "myclirc")
+    default_config_file = resources.files(mycli).joinpath("myclirc").open('r')
     return read_config_file(default_config_file, list_values=list_values)
 
 
 def write_default_config(destination: str, overwrite: bool = False) -> None:
     import mycli
 
-    default_config = resources.read_text(mycli, "myclirc")
+    with resources.files(mycli).joinpath("myclirc").open('r') as f:
+        default_config = f.read()
     destination = os.path.expanduser(destination)
     if not overwrite and exists(destination):
         return
@@ -177,7 +188,7 @@ def encrypt_mylogin_cnf(plaintext: IO[str]) -> BytesIO:
         return bytes(rkey)
 
     def encode_line(plaintext: str, real_key: bytes, buf_len: int) -> bytes:
-        aes = pyaes.AESModeOfOperationECB(real_key)
+        aes = AES.new(real_key, AES.MODE_ECB)
         text_len = len(plaintext)
         pad_len = buf_len - text_len
         pad_chr = bytes(chr(pad_len), "utf8")
@@ -252,7 +263,7 @@ def read_and_decrypt_mylogin_cnf(f: BinaryIO) -> BytesIO | None:
 
     # Create a bytes buffer to hold the plaintext.
     plaintext = BytesIO()
-    aes = pyaes.AESModeOfOperationECB(rkey_b)
+    aes = AES.new(rkey_b, AES.MODE_ECB)
 
     while True:
         # Read the length of the ciphertext.
@@ -291,7 +302,7 @@ def str_to_bool(s: str | bool) -> bool:
     elif s.lower() in false_values:
         return False
     else:
-        raise ValueError("not a recognized boolean value: {0}".format(s))
+        raise ValueError(f'not a recognized boolean value: {s}')
 
 
 def strip_matching_quotes(s: str) -> str:
