@@ -1,6 +1,7 @@
 # type: ignore
 
 from collections import namedtuple
+import csv
 import os
 import shutil
 from tempfile import NamedTemporaryFile
@@ -8,11 +9,14 @@ from textwrap import dedent
 
 import click
 from click.testing import CliRunner
+from pymysql.err import OperationalError
 
 from mycli.main import MyCli, cli, thanks_picker
+from mycli.packages.parseutils import is_valid_connection_scheme
+import mycli.packages.special
 from mycli.packages.special.main import COMMANDS as SPECIAL_COMMANDS
-from mycli.sqlexecute import ServerInfo
-from test.utils import HOST, PASSWORD, PORT, USER, dbtest, run
+from mycli.sqlexecute import ServerInfo, SQLExecute
+from test.utils import DATABASE, HOST, PASSWORD, PORT, USER, dbtest, run
 
 test_dir = os.path.abspath(os.path.dirname(__file__))
 project_dir = os.path.dirname(test_dir)
@@ -37,6 +41,317 @@ CLI_ARGS = [
 ]
 
 
+@dbtest
+def test_binary_display_hex(executor, capsys):
+    m = MyCli()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    m.explicit_pager = False
+    sqlresult = next(m.sqlexecute.run("select b'01101010' AS binary_test"))
+    formatted = m.format_output(
+        sqlresult.title,
+        sqlresult.results,
+        sqlresult.headers,
+        False,
+        False,
+        "<nope>",
+        "right",
+        "hex",
+        None,
+    )
+    m.output(formatted, sqlresult.status)
+    expected = " 0x6a "
+    stdout = capsys.readouterr().out
+    assert expected in stdout
+
+
+@dbtest
+def test_binary_display_utf8(executor, capsys):
+    m = MyCli()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    m.explicit_pager = False
+    sqlresult = next(m.sqlexecute.run("select b'01101010' AS binary_test"))
+    formatted = m.format_output(
+        sqlresult.title,
+        sqlresult.results,
+        sqlresult.headers,
+        False,
+        False,
+        "<nope>",
+        "right",
+        "utf8",
+        None,
+    )
+    m.output(formatted, sqlresult.status)
+    expected = " j "
+    stdout = capsys.readouterr().out
+    assert expected in stdout
+
+
+@dbtest
+def test_select_from_empty_table(executor):
+    run(executor, """create table t1(id int)""")
+    sql = "select * from t1"
+    runner = CliRunner()
+    result = runner.invoke(cli, args=CLI_ARGS + ["-t"], input=sql)
+    expected = dedent("""\
+        +----+
+        | id |
+        +----+
+        +----+""")
+    assert expected in result.output
+
+
+def test_is_valid_connection_scheme_valid(executor, capsys):
+    is_valid, scheme = is_valid_connection_scheme("mysql://test@localhost:3306/dev")
+    assert is_valid
+
+
+def test_is_valid_connection_scheme_invalid(executor, capsys):
+    is_valid, scheme = is_valid_connection_scheme("nope://test@localhost:3306/dev")
+    assert not is_valid
+
+
+@dbtest
+def test_ssl_mode_on(executor, capsys):
+    runner = CliRunner()
+    ssl_mode = "on"
+    sql = "select * from performance_schema.session_status where variable_name = 'Ssl_cipher'"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--csv", "--ssl-mode", ssl_mode], input=sql)
+    result_dict = next(csv.DictReader(result.stdout.split("\n")))
+    ssl_cipher = result_dict.get("VARIABLE_VALUE", None)
+    assert ssl_cipher
+
+
+@dbtest
+def test_ssl_mode_auto(executor, capsys):
+    runner = CliRunner()
+    ssl_mode = "auto"
+    sql = "select * from performance_schema.session_status where variable_name = 'Ssl_cipher'"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--csv", "--ssl-mode", ssl_mode], input=sql)
+    result_dict = next(csv.DictReader(result.stdout.split("\n")))
+    ssl_cipher = result_dict.get("VARIABLE_VALUE", None)
+    assert ssl_cipher
+
+
+@dbtest
+def test_ssl_mode_off(executor, capsys):
+    runner = CliRunner()
+    ssl_mode = "off"
+    sql = "select * from performance_schema.session_status where variable_name = 'Ssl_cipher'"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--csv", "--ssl-mode", ssl_mode], input=sql)
+    result_dict = next(csv.DictReader(result.stdout.split("\n")))
+    ssl_cipher = result_dict.get("VARIABLE_VALUE", None)
+    assert not ssl_cipher
+
+
+@dbtest
+def test_ssl_mode_overrides_ssl(executor, capsys):
+    runner = CliRunner()
+    ssl_mode = "off"
+    sql = "select * from performance_schema.session_status where variable_name = 'Ssl_cipher'"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--csv", "--ssl-mode", ssl_mode, "--ssl"], input=sql)
+    result_dict = next(csv.DictReader(result.stdout.split("\n")))
+    ssl_cipher = result_dict.get("VARIABLE_VALUE", None)
+    assert not ssl_cipher
+
+
+@dbtest
+def test_ssl_mode_overrides_no_ssl(executor, capsys):
+    runner = CliRunner()
+    ssl_mode = "on"
+    sql = "select * from performance_schema.session_status where variable_name = 'Ssl_cipher'"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--csv", "--ssl-mode", ssl_mode, "--no-ssl"], input=sql)
+    result_dict = next(csv.DictReader(result.stdout.split("\n")))
+    ssl_cipher = result_dict.get("VARIABLE_VALUE", None)
+    assert ssl_cipher
+
+
+@dbtest
+def test_reconnect_database_is_selected(executor, capsys):
+    m = MyCli()
+    m.register_special_commands()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    try:
+        next(m.sqlexecute.run(f"use {DATABASE}"))
+        next(m.sqlexecute.run(f"kill {m.sqlexecute.connection_id}"))
+    except OperationalError:
+        pass  # expected as the connection was killed
+    except Exception as e:
+        raise e
+    m.reconnect()
+    try:
+        next(m.sqlexecute.run("show tables")).results.fetchall()
+    except Exception as e:
+        raise e
+
+
+@dbtest
+def test_reconnect_no_database(executor, capsys):
+    m = MyCli()
+    m.register_special_commands()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    sql = "\\r"
+    result = next(mycli.packages.special.execute(executor, sql))
+    stdout, _stderr = capsys.readouterr()
+    assert result.status is None
+    assert "Already connected" in stdout
+
+
+@dbtest
+def test_reconnect_with_different_database(executor):
+    m = MyCli()
+    m.register_special_commands()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    database_1 = "mycli_test_db"
+    database_2 = "mysql"
+    sql_1 = f"use {database_1}"
+    sql_2 = f"\\r {database_2}"
+    _result_1 = next(mycli.packages.special.execute(executor, sql_1))
+    result_2 = next(mycli.packages.special.execute(executor, sql_2))
+    expected = f'You are now connected to database "{database_2}" as user "{USER}"'
+    assert expected in result_2.status
+
+
+@dbtest
+def test_reconnect_with_same_database(executor):
+    m = MyCli()
+    m.register_special_commands()
+    m.sqlexecute = SQLExecute(
+        None,
+        USER,
+        PASSWORD,
+        HOST,
+        PORT,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+        None,
+    )
+    database = "mysql"
+    sql = f"\\u {database}"
+    result = next(mycli.packages.special.execute(executor, sql))
+    sql = f"\\r {database}"
+    result = next(mycli.packages.special.execute(executor, sql))
+    expected = f'You are already connected to database "{database}" as user "{USER}"'
+    assert expected in result.status
+
+
+@dbtest
+def test_prompt_no_host_only_socket(executor):
+    mycli = MyCli()
+    mycli.prompt_format = "\\t \\u@\\h:\\d> "
+    mycli.sqlexecute = SQLExecute
+    mycli.sqlexecute.server_info = ServerInfo.from_version_string("8.0.44-0ubuntu0.24.04.1")
+    mycli.sqlexecute.host = None
+    mycli.sqlexecute.socket = "/var/run/mysqld/mysqld.sock"
+    mycli.sqlexecute.user = "root"
+    mycli.sqlexecute.dbname = "mysql"
+    mycli.sqlexecute.port = "3306"
+    prompt = mycli.get_prompt(mycli.prompt_format)
+    assert prompt == "MySQL root@localhost:mysql> "
+
+
+@dbtest
+def test_prompt_socket_overrides_port(executor):
+    mycli = MyCli()
+    mycli.prompt_format = "\\t \\u@\\h:\\k \\d> "
+    mycli.sqlexecute = SQLExecute
+    mycli.sqlexecute.server_info = ServerInfo.from_version_string("8.0.44-0ubuntu0.24.04.1")
+    mycli.sqlexecute.host = None
+    mycli.sqlexecute.socket = "/var/run/mysqld/mysqld.sock"
+    mycli.sqlexecute.user = "root"
+    mycli.sqlexecute.dbname = "mysql"
+    mycli.sqlexecute.port = "3306"
+    prompt = mycli.get_prompt(mycli.prompt_format)
+    assert prompt == "MySQL root@localhost:mysqld.sock mysql> "
+
+
 @dbtest
 def test_enable_show_warnings(executor):
     mycli = MyCli()
@@ -55,6 +370,17 @@ def test_disable_show_warnings(executor):
     assert result[0]["status"] == "Show warnings disabled."
 
 
+@dbtest
+def test_output_ddl_with_warning_and_show_warnings_enabled(executor):
+    runner = CliRunner()
+    db = "mycli_test_db"
+    table = "table_that_definitely_does_not_exist_1234"
+    sql = f"DROP TABLE IF EXISTS {db}.{table}"
+    result = runner.invoke(cli, args=CLI_ARGS + ["--show-warnings", "--no-warn"], input=sql)
+    expected = "Level\tCode\tMessage\nNote\t1051\tUnknown table 'mycli_test_db.table_that_definitely_does_not_exist_1234'\n"
+    assert expected in result.output
+
+
 @dbtest
 def test_output_with_warning_and_show_warnings_enabled(executor):
     runner = CliRunner()
@@ -128,6 +454,42 @@ def test_execute_arg(executor):
     assert expected in result.output
 
 
+@dbtest
+def test_execute_arg_with_checkpoint(executor):
+    run(executor, "create table test (a text)")
+    run(executor, 'insert into test values("abc")')
+
+    sql = "select * from test;"
+    runner = CliRunner()
+
+    with NamedTemporaryFile(mode="w", delete=False) as checkpoint:
+        checkpoint.close()
+
+    result = runner.invoke(cli, args=CLI_ARGS + ["--execute", sql, f"--checkpoint={checkpoint.name}"])
+    assert result.exit_code == 0
+
+    with open(checkpoint.name, 'r') as f:
+        contents = f.read()
+    assert sql in contents
+    os.remove(checkpoint.name)
+
+    sql = 'select 10 from nonexistent_table;'
+    result = runner.invoke(cli, args=CLI_ARGS + ["--execute", sql, f"--checkpoint={checkpoint.name}"])
+    assert result.exit_code != 0
+
+    with open(checkpoint.name, 'r') as f:
+        contents = f.read()
+    assert sql not in contents
+
+    # delete=False means we should try to clean up
+    # we don't really need "try" here as open() would have already failed
+    try:
+        if os.path.exists(checkpoint.name):
+            os.remove(checkpoint.name)
+    except Exception as e:
+        print(f"An error occurred while attempting to delete the file: {e}")
+
+
 @dbtest
 def test_execute_arg_with_table(executor):
     run(executor, "create table test (a text)")
@@ -184,7 +546,7 @@ def test_batch_mode_table(executor):
         +----------+
         | count(*) |
         +----------+
-        | 3        |
+        |        3 |
         +----------+
         +-----+
         | a   |
@@ -249,6 +611,7 @@ def output(monkeypatch, terminal_size, testdata, explicit_pager, expect_pager):
         dbname = "test"
         server_info = ServerInfo.from_version_string("unknown")
         port = 0
+        socket = ''
 
         def server_type(self):
             return ["test"]
@@ -306,7 +669,9 @@ def test_reserved_space_is_integer(monkeypatch):
         assert isinstance(mycli.get_reserved_space(), int)
 
 
-def test_list_dsn():
+def test_list_dsn(monkeypatch):
+    monkeypatch.setattr(MyCli, "system_config_files", [])
+    monkeypatch.setattr(MyCli, "pwd_config_file", os.path.join(test_dir, "does_not_exist.myclirc"))
     runner = CliRunner()
     # keep Windows from locking the file with delete=False
     with NamedTemporaryFile(mode="w", delete=False) as myclirc:
@@ -386,13 +751,18 @@ def test_dsn(monkeypatch):
             pass
 
     class MockMyCli:
-        config = {"alias_dsn": {}}
+        config = {
+            "main": {},
+            "alias_dsn": {},
+        }
 
         def __init__(self, **args):
             self.logger = Logger()
             self.destructive_warning = False
             self.main_formatter = Formatter()
             self.redirect_formatter = Formatter()
+            self.ssl_mode = "auto"
+            self.my_cnf = {"client": {}, "mysqld": {}}
 
         def connect(self, **args):
             MockMyCli.connect_args = args
@@ -447,7 +817,10 @@ def test_dsn(monkeypatch):
         and MockMyCli.connect_args["database"] == "arg_database"
     )
 
-    MockMyCli.config = {"alias_dsn": {"test": "mysql://alias_dsn_user:alias_dsn_passwd@alias_dsn_host:4/alias_dsn_database"}}
+    MockMyCli.config = {
+        "main": {},
+        "alias_dsn": {"test": "mysql://alias_dsn_user:alias_dsn_passwd@alias_dsn_host:4/alias_dsn_database"},
+    }
     MockMyCli.connect_args = None
 
     # When a user uses a DSN from the configuration file (alias_dsn),
@@ -462,7 +835,10 @@ def test_dsn(monkeypatch):
         and MockMyCli.connect_args["database"] == "alias_dsn_database"
     )
 
-    MockMyCli.config = {"alias_dsn": {"test": "mysql://alias_dsn_user:alias_dsn_passwd@alias_dsn_host:4/alias_dsn_database"}}
+    MockMyCli.config = {
+        "main": {},
+        "alias_dsn": {"test": "mysql://alias_dsn_user:alias_dsn_passwd@alias_dsn_host:4/alias_dsn_database"},
+    }
     MockMyCli.connect_args = None
 
     # When a user uses a DSN from the configuration file (alias_dsn)
@@ -550,13 +926,18 @@ def test_ssh_config(monkeypatch):
             pass
 
     class MockMyCli:
-        config = {"alias_dsn": {}}
+        config = {
+            "main": {},
+            "alias_dsn": {},
+        }
 
         def __init__(self, **args):
             self.logger = Logger()
             self.destructive_warning = False
             self.main_formatter = Formatter()
             self.redirect_formatter = Formatter()
+            self.ssl_mode = "auto"
+            self.my_cnf = {"client": {}, "mysqld": {}}
 
         def connect(self, **args):
             MockMyCli.connect_args = args
@@ -666,3 +1047,22 @@ def test_global_init_commands(executor):
     expected = "sql_select_limit\t9999\n"
     assert result.exit_code == 0
     assert expected in result.output
+
+
+@dbtest
+def test_execute_with_logfile(executor):
+    """Test that --execute combines with --logfile"""
+    sql = 'select 1'
+    runner = CliRunner()
+
+    with NamedTemporaryFile(mode="w", delete=False) as logfile:
+        result = runner.invoke(mycli.main.cli, args=CLI_ARGS + ["--logfile", logfile.name, "--execute", sql])
+        assert result.exit_code == 0
+
+    assert os.path.getsize(logfile.name) > 0
+
+    try:
+        if os.path.exists(logfile.name):
+            os.remove(logfile.name)
+    except Exception as e:
+        print(f"An error occurred while attempting to delete the file: {e}")
